; Scheme_1 compiler/interpreter (mutable variables)
; By Nils M Holm, 2016, 2018
;
; Provided under the Creative Common Zero (CC0) license
; (https://creativecommons.org/publicdomain/zero/1.0/)
;
; From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/

; Examples: (comp '(((lambda (x) (lambda (y) (cons x y))) 'x) 'y))
;           (evcomp '(((lambda (x) (lambda (y) (cons x y))) 'x) 'y))

(define version 3)

(define (but-last x)
  (reverse (cdr (reverse x))))

(define (last x)
  (car (reverse x)))

; Compiler

(define (comp x)

  (define (bind x e n)
    (cons (cons x n) e))

  (define (cconv x e n)
    (cond ((assq x e)
            => (lambda (b)
                 `(%ref ,(cdr b))))
          ((not (pair? x))
            x)
          ((eq? 'quote (car x))
            x)
          ((eq? 'lambda (car x))
            `(lambda ,n ,@(cconv (cddr x)
                                 (bind (caadr x) e n)
                                 (+ 1 n))))
          (else
            (cons (cconv (car x) e n)
                  (cconv (cdr x) e n)))))

  (define (closure-conv x)
    (cconv x '() 0))

  (define prim-op   cadr)
  (define prim-narg caddr)

  (define primitives '((cons  %cons  2)
                       (car   %car   1)
                       (cdr   %cdr   1)
                       (pair? %pair? 1)
                       (null? %null? 1)))

  (define (new-sym i)
    (string->symbol (string-append "x" (number->string i))))

  (define (prim-fun x)
    (let vars ((i (prim-narg x))
               (v '()))
      (if (zero? i)
          (cons '%lambda
                (cdr (closure-conv
                       `(lambda ,v (,(prim-op x) ,@v)))))
          (vars (- i 1) (cons (new-sym i) v)))))

  (define (comp-expr x t)
    (cond ((assq x primitives)
            => prim-fun)
          ((not (pair? x))
            x)
          ((eq? 'quote (car x))
            `(%quote ,(cadr x)))
          ((eq? 'lambda (car x))
            `(%lambda ,(cadr x)
                      ,(if (null? (cdddr x))
                           (comp-expr (caddr x) #t)
                           (comp-expr (cons 'begin (cddr x)) #t))))
          ((eq? '%ref (car x))
            x)
          ((eq? 'if (car x))
            `(%if ,(comp-expr (cadr x) #f)
                  ,(comp-expr (caddr x) t)
                  ,(comp-expr (cadddr x) t)))
          ((eq? 'set! (car x))
            `(%set! ,(cadr x) ,(comp-expr (caddr x) #f)))
          ((eq? 'begin (car x))
            `(%begin ,@(map (lambda (x)
                              (comp-expr x #f))
                            (but-last (cdr x)))
                     ,(comp-expr (last x) t)))
          ((assq (car x) primitives)
            => (lambda (i)
                 `(,(prim-op i)
                   ,@(map (lambda (x)
                            (comp-expr x #f))
                          (cdr x)))))
          (else
            `(,(if t '%tail-apply '%apply)
              ,@(map (lambda (x)
                       (comp-expr x #f))
                     x)))))
  
  (comp-expr (closure-conv x) #f))

; Interpreter / Abstract Machine

(define (eval x)

  (define E '#())
  (define S '())

  (define (extend-vector k e)
    (let ((v (make-vector (+ 1 k))))
      (let loop ((i 0))
        (if (= i k)
            v
            (begin (vector-set! v i (vector-ref e i))
                   (loop (+ 1 i)))))))

  (define (make-fun x)
    `(%fn ,(cadr x) ,(extend-vector (cadr x) E) ,(caddr x)))

  (define fun-slot cadr)
  (define fun-env  caddr)
  (define fun-term cadddr)

  (define (push x)
    (set! S (cons x S)))

  (define (cleanup)
    (vector-set!
      E
      (- (vector-length E) 1)
      (caddr S))
    (set! E (cadr S))
    (set! S (cdddr S)))

  (define fun cadr)
  (define arg caddr)

  (define box list)
  (define unbox car)

  (define (ev-apply x)
    (let ((fn (ev (fun x))))
      (push (vector-ref (fun-env fn) (fun-slot fn)))
      (vector-set! (fun-env fn) (fun-slot fn) (box (ev (arg x))))
      (push E)
      (push 'C)
      (set! E (fun-env fn))
      (let ((v (ev (fun-term fn))))
        (cleanup)
        v)))

  (define (ev-tail-apply x)
    (let ((fn (ev (fun x)))
          (a  (box (ev (arg x)))))
      (cleanup)
      (push (vector-ref (fun-env fn) (fun-slot fn)))
      (vector-set! (fun-env fn) (fun-slot fn) a)
      (push E)
      (push 'C)
      (set! E (fun-env fn))
      (ev (fun-term fn))))

  (define (ev-set n x)
    (set-car! (vector-ref E (cadr n)) x)
    x)

  (define (ev x)
    (let ((f (car x)))
      (cond ((eq? '%quote f)
              (cadr x))
            ((eq? '%lambda f)
              (make-fun x))
            ((eq? '%apply f)
              (ev-apply x))
            ((eq? '%tail-apply f)
              (ev-tail-apply x))
            ((eq? '%ref f)
              (unbox (vector-ref E (cadr x))))
            ((eq? '%if f)
              (if (null? (ev (cadr x)))
                  (ev (cadddr x))
                  (ev (caddr x))))
            ((eq? '%set! f)
              (ev-set (cadr x) (ev (caddr x))))
            ((eq? '%begin f)
              (for-each ev (but-last (cdr x)))
              (ev (last x)))
            ((eq? '%cons f)
              (cons (ev (cadr x)) (ev (caddr x))))
            ((eq? '%car f)
              (car (ev (cadr x))))
            ((eq? '%cdr f)
              (cdr (ev (cadr x))))
            ((eq? '%pair? f)
              (if (pair? (ev (cadr x))) 't '()))
            ((eq? '%null? f)
              (if (null? (ev (cadr x))) 't '()))
            (else
              (error "lc: cannot eval" x)))))

  (ev x))

(define (evcomp x)
  (eval (comp x)))
