; Scheme_3 compiler/interpreter, loop unrolling, low-level VM
; By Nils M Holm, 2016, 2018
;
; Provided under the Creative Common Zero (CC0) license
; (https://creativecommons.org/publicdomain/zero/1.0/)
;
; From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/

; Examples: (comp '(((lambda (x) (lambda (y) (cons x y))) 'x) 'y))
;           (evcomp '(((lambda (x) (lambda (y) (cons x y))) 'x) 'y))

(define version 7)

(define (but-last x)
  (reverse (cdr (reverse x))))

(define (last x)
  (car (reverse x)))

(define (lc-error m . a)
  (apply error (string-append "lc: " m) a))

; Compiler

(define prim-op   cadr)
(define prim-args cddr)
(define prim-narg caddr)

(define primitives '((cons  %cons  2 #f)
                     (car   %car   1 #f)
                     (cdr   %cdr   1 #f)
                     (pair? %pair? 1 #f)
                     (null? %null? 1 #f)))

(define (stage1 x)

  (define (flatten x)
    (define (f x r)
      (cond ((null? x)
              r)
            ((pair? x)
              (f (car x)
                 (f (cdr x) r)))
            (else
              (cons x r))))
    (f x '()))

  (define (bind xs e n)
    (if (null? xs)
        e
        (cons (cons (car xs) n)
              (bind (cdr xs) e (+ 1 n)))))

  (define (cconv x e n)
    (cond ((assq x e)
            => (lambda (b)
                 `(%ref ,(cdr b))))
          ((not (pair? x))
            x)
          ((eq? 'quote (car x))
            x)
          ((eq? 'lambda (car x))
            (let ((a (flatten (cadr x))))
              `(lambda (,n
                        ,(+ n (length a))
                        ,(not (list? (cadr x))))
                       ,@(cconv (cddr x)
                                (bind a e n)
                                (+ (length a) n)))))
          (else
            (cons (cconv (car x) e n)
                  (cconv (cdr x) e n)))))

  (define (closure-conv x)
    (cconv x '() 0))

  (define (assert-args! x n v)
    (let ((k (length (cdr x))))
      (cond ((= k n))
            ((and (> k n) v))
            (else (lc-error "wrong argument count" x)))))

  (define (new-sym i)
    (string->symbol
      (string-append
        "x"
        (number->string i))))

  (define (prim-fun x)
    (let vars ((i (prim-narg x))
               (v '()))
      (if (zero? i)
          (cons '%lambda
                (cdr (closure-conv
                       `(lambda ,v (,(prim-op x) ,@v)))))
          (vars (- i 1) (cons (new-sym i) v)))))

  (define (comp-expr x t)
    (cond ((assq x primitives)
            => prim-fun)
          ((symbol? x)
            (lc-error "undefined" x))
          ((not (pair? x))
            x)
          ((eq? 'quote (car x))
            (assert-args! x 1 #f)
            `(%quote ,(cadr x)))
          ((eq? 'lambda (car x))
            (assert-args! x 2 #t)
            `(%lambda ,(cadr x)
                      ,(if (null? (cdddr x))
                           (comp-expr (caddr x) #t)
                           (comp-expr (cons 'begin (cddr x))
                                      #t))))
          ((eq? '%ref (car x))
            x)
          ((eq? 'if (car x))
            (assert-args! x 3 #f)
            `(%if ,(comp-expr (cadr x) #f)
                  ,(comp-expr (caddr x) t)
                  ,(comp-expr (cadddr x) t)))
          ((eq? 'set! (car x))
            (assert-args! x 2 #f)
            `(%set! ,(cadr x) ,(comp-expr (caddr x) #f)))
          ((eq? 'begin (car x))
            (assert-args! x 1 #t)
            `(%begin ,@(map (lambda (x)
                              (comp-expr x #f))
                            (but-last (cdr x)))
                     ,(comp-expr (last x) t)))
          ((assq (car x) primitives)
            => (lambda (i)
                 (apply assert-args! x (prim-args i))
                 `(,(prim-op i)
                   ,@(map (lambda (x)
                            (comp-expr x #f))
                          (cdr x)))))
          (else
            `(,(if t '%tail-apply '%apply)
              ,@(map (lambda (x)
                       (comp-expr x #f))
                     x)))))
  
  (comp-expr (closure-conv x) #f))

(define (stage2 x)

  (define fp-osize car)
  (define fp-nsize cadr)
  (define fp-var?  caddr)

  (define Prog '())
  (define Lab  0)

  (define Lit '(()))
  (define Nlit 0)

  (define (newlabel)
    (set! Lab (+ 1 Lab))
    Lab)

  (define (newlit)
    (set! Nlit (+ 1 Nlit))
    Nlit)

  (define (emit x)
    (set! Prog (cons x Prog)))

  (define (comp-lambda x)
    (let ((osize  (fp-osize (cadr x)))
          (nsize  (fp-nsize (cadr x)))
          (funlab (newlabel))
          (endlab (newlabel)))
      (emit `(%%jump ,endlab))
      (emit `(%%label ,funlab))
      (if (fp-var? (cadr x))
          (emit `(%%collect-args ,(- nsize osize))))
      (emit `(%%save-vars ,osize ,nsize))
      (emit `(%%assert-args ,(- nsize osize)))
      (let loop ((i 0))
        (if (< i (- nsize osize))
            (begin (emit `(%%setup-env ,i ,(- nsize i 1)))
                   (loop (+ 1 i)))))
      (comp-s2 (caddr x))
      (emit `(%%return))
      (emit `(%%label ,endlab))
      (emit `(%%fun ,funlab ,osize ,nsize))))

  (define (comp-apply x t)
    (for-each comp-s2 (cddr x))
    (emit `(%%val ,(length (cddr x))))
    (comp-s2 (cadr x))
    (emit (if t '(%%tail-apply) '(%%apply))))

  (define (comp-ref x)
    (emit `(%%ref ,(cadr x))))

  (define (comp-quote x)
    (if (null? (cadr x))
        (emit '(%%lit 0))
        (begin (set! Lit (cons (cadr x) Lit))
               (emit `(%%lit ,(newlit))))))

  (define (comp-if x)
    (let ((tlab (newlabel))
          (flab (newlabel)))
      (comp-s2 (cadr x))
      (emit `(%%jump-false ,flab))
      (comp-s2 (caddr x))
      (emit `(%%jump ,tlab))
      (emit `(%%label ,flab))
      (comp-s2 (cadddr x))
      (emit `(%%label ,tlab))))

  (define (comp-begin x)
    (let loop ((x (cdr x)))
      (cond ((null? (cdr x))
              (comp-s2 (car x)))
            (else
              (comp-s2 (car x))
              (emit '(%%drop))
              (loop (cdr x))))))

  (define (comp-set x)
    (comp-s2 (caddr x))
    (let ((loc (cadr x)))
      (emit `(%%set-ref! ,(cadr loc)))))

  (define (comp-prim x p)
    (for-each comp-s2 (cdr x))
    (emit (cond ((eq? '%cons  p) '(%%cons))
                ((eq? '%car   p) '(%%car))
                ((eq? '%cdr   p) '(%%cdr))
                ((eq? '%pair? p) '(%%pair?))
                ((eq? '%null? p) '(%%null?)))))

  (define prim-ops (map cadr primitives))

  (define (comp-s2 x)
    (cond ((eq? '%lambda (car x)) (comp-lambda x))
          ((eq? '%apply (car x)) (comp-apply x #f))
          ((eq? '%tail-apply (car x)) (comp-apply x #t))
          ((eq? '%ref (car x)) (comp-ref x))
          ((eq? '%quote (car x)) (comp-quote x))
          ((eq? '%if (car x)) (comp-if x))
          ((eq? '%set! (car x)) (comp-set x))
          ((eq? '%begin (car x)) (comp-begin x))
          ((memq (car x) prim-ops)
            => (lambda (i)
                 (comp-prim x (car i))))
          (else (error "unknown op in stage2" x))))

  (comp-s2 x)
  (emit '(%%halt))
  (cons (cons '%%litpool (reverse Lit))
        (reverse Prog)))

(define (comp x)
  (stage2 (stage1 x)))

; Virtual Machine

(define (vmrun x)

  (define Frame-saved 1)
  (define Frame-env   2)
  (define Frame-ip    3)
  (define Frame-narg  4)
  (define Frame-size  4)
  (define Frame-skip  5)

  (define Litpool #())

  (define Stack (make-vector 1000))
  (define Stack-len 0)
  (define Sp 0)

  (define Prog #())
  (define Ip 0)

  (define Addrs #())

  (define Env #())

  (define box list)
  (define unbox car)

  (define (resolve x)
    (let ((t (make-vector (length x))))
      (let loop ((x x)
                 (i 0))
        (cond ((null? x)
                t)
              ((eq? '%%label (caar x))
                (vector-set! t (cadar x) i)
                (loop (cdr x) (+ 1 i)))
              (else
                (loop (cdr x) (+ 1 i)))))))

  (define (stack-ref i)
    (vector-ref Stack (- Sp i)))

  (define (stack-set! i x)
    (vector-set! Stack (- Sp i) x))

  (define (copy-vector! d s n k)
    (let loop ((i 0))
      (if (< i k)
          (begin (vector-set! d (+ n i) (vector-ref s i))
                 (loop (+ 1 i))))))

  (define (push x)
    (if (= Sp Stack-len)
        (let ((k (+ Stack-len 1000)))
          (let ((s (make-vector k)))
            (copy-vector! s Stack 0 Stack-len)
            (set! Stack s)
            (set! Stack-len k))))
    (stack-set! 0 x)
    (set! Sp (+ 1 Sp)))

  (define (pop)
    (set! Sp (- Sp 1))
    (stack-ref 0))

  (define (extend n m)
    (let ((e (make-vector m))
          (k (- m 1)))
      (let loop ((i (- n 1)))
        (if (not (negative? i))
            (begin (vector-set! e i (vector-ref Env i))
                   (loop (- i 1)))
            e))))

  (define (collect-args n)
    (let* ((s-env (pop))
           (s-ip  (pop))
           (na    (pop)))
      (if (> n (+ 1 na))
          (lc-error "too few arguments"))
      (let loop ((i (- n 1))
                 (a '()))
        (if (< i na)
            (loop (+ 1 i) (cons (pop) a))
            (begin (push a)
                   (push n)
                   (push s-ip)
                   (push s-env))))))

  (define (copy-vector! d s n k)
    (let loop ((i 0)) 
      (if (< i k)
          (begin (vector-set! d (+ n i) (vector-ref s i))
                 (loop (+ 1 i))))))

  (define (sub-vector v n k)
    (let ((nv (make-vector (- k n))))
      (let loop ((i n))
        (if (= i k)
            nv
            (begin (vector-set! nv (- i n) (vector-ref v i))
                   (loop (+ 1 i)))))))

  (define (save-vars n k)
    (push (sub-vector Env n k)))

  (define (setup-env i j)
    (let ((b (box (stack-ref (+ Frame-skip i)))))
      (vector-set! Env j b)))

  (define (restore-vars sv)
    (let ((k (vector-length sv)))
      (copy-vector! Env sv (- (vector-length Env) k) k)))

  (define (move-frame d s k)
    (let loop ((d d) (s s) (k k))
      (if (not (zero? k))
          (begin (vector-set!
                   Stack
                   d
                   (vector-ref Stack s))
                 (loop (- d 1) (- s 1) (- k 1))))))

  (define (tail-call)
    (let* ((n (stack-ref 1))                 ; narg
           (m (stack-ref (+ Frame-skip n)))) ; caller's narg
      (restore-vars (stack-ref (+ 2 n)))
      (cond ((= n m)
              (move-frame (- Sp n Frame-skip)
                          (- Sp 1)
                          (+ 1 n))
              (set! Sp (- Sp n 2)))
            (else
              (let ((s-ip  (stack-ref (+ n 1 Frame-ip)))
                    (s-env (stack-ref (+ n 1 Frame-env))))
                (move-frame (- Sp m Frame-skip)
                            (- Sp 1)
                            (+ 1 n))
                (set! Sp (+ 1 (- Sp m Frame-skip)))
                (push s-ip)
                (push s-env))))))

  (define (assert-fn! x)
    (if (or (not (pair? x))
            (not (eq? '%%fn (car x))))
      (lc-error "application of non-function")))

  (define (assert-args! x)
    (if (not (= x (stack-ref Frame-narg)))
        (lc-error "wrong number of arguments")))

  (define (dispatch x)
    (let ((op (car x)))
      (cond ((eq? op '%%apply)
              (let ((f (pop)))
                (assert-fn! f)
                (push Ip)
                (push Env)
                (set! Env (caddr f))
                (set! Ip (cadr f))))
            ((eq? op '%%tail-apply)
              (let ((f (pop)))
                (assert-fn! f)
                (tail-call)
                (set! Env (caddr f))
                (set! Ip (cadr f))))
            ((eq? op '%%assert-args)
              (assert-args! (cadr x)))
            ((eq? op '%%car)
              (push (car (pop))))
            ((eq? op '%%cdr)
              (push (cdr (pop))))
            ((eq? op '%%collect-args)
              (collect-args (cadr x)))
            ((eq? op '%%cons)
              (let ((y (pop)))
                (push (cons (pop) y))))
            ((eq? op '%%drop)
              (pop))
            ((eq? op '%%save-vars)
              (save-vars (cadr x) (caddr x)))
            ((eq? op '%%setup-env)
              (setup-env (cadr x) (caddr x)))
            ((eq? op '%%fun)
              (let ((e (extend (caddr x) (cadddr x))))
                (push (list '%%fn (vector-ref Addrs (cadr x)) e))))
            ((eq? op '%%jump)
              (set! Ip (vector-ref Addrs (cadr x))))
            ((eq? op '%%jump-false)
              (if (null? (pop))
                  (set! Ip (vector-ref Addrs (cadr x)))))
            ((eq? op '%%label))
            ((eq? op '%%lit)
              (push (vector-ref Litpool (cadr x))))
            ((eq? op '%%litpool)
              (set! Litpool (list->vector (cdr x))))
            ((eq? op '%%val)
              (push (cadr x)))
            ((eq? op '%%pair?)
              (push (if (pair? (pop)) 't '())))
            ((eq? op '%%null?)
              (push (if (null? (pop)) 't '())))
            ((eq? op '%%ref)
              (push (unbox (vector-ref Env (cadr x)))))
            ((eq? op '%%return)
              (let ((v (pop)))
                (restore-vars (pop))
                (set! Env (pop))
                (set! Ip (pop))
                (set! Sp (- Sp (pop) 1))
                (push v)))
            ((eq? op '%%set-ref!)
              (let ((v (pop)))
                (set-car! (vector-ref Env (cadr x)) v)
                (push v))))))

  (define (run)
    (let ((i (vector-ref Prog Ip)))
      (if (not (eq? '%%halt (car i)))
          (begin (dispatch i)
                 (set! Ip (+ 1 Ip))
                 (run)))))

  (set! Stack-len (vector-length Stack))
  (set! Prog (list->vector x))
  (set! Addrs (resolve x))
  (run)
  (if (not (zero? Sp))
      (pop)))

(define (evcomp x)
  (vmrun (comp x)))
